;; -*- no-byte-compile: t; lexical-binding: nil -*-
(define-package "ranger" "20260527.2231"
  "Make dired more like ranger."
  '((emacs "27.1"))
  :url "https://github.com/ralesi/ranger"
  :commit "cfcab21cd21eb48709fb4436ba84f3e2a048a432"
  :revdesc "cfcab21cd21e"
  :keywords '("files" "convenience" "dired")
  :authors '(("Rich Alesi" . "https://github.com/ralesi"))
  :maintainers '(("Rich Alesi" . "https://github.com/ralesi")))
